﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace TasksAsFutures {
	class Program {
		static Task<int> GetValueAsync(int x) {		// since .NET 4 (scheduled into TaskScheduler.Current)
			return Task<int>.Factory.StartNew(() => x);
		}

		static Task<int> GetValueAsync45(int x) {   // only in .NET 4.5 (scheduled into TaskScheduler.Default)
			return Task.Run(() => x);
		}

		static Task<int> IncAsync(Task<int> v) {
			return Task.Run(() => {
				return v.Result + 1;
			});
		}

		static Task<int> Inc2Async(Task<int> v) {
			return Task.Run(() => {
				v.Wait();
				return v.Result + 1;
			});
		}

		static Task<int> IncAlmostOptimalAsync(Task<int> v) {
			return v.ContinueWith(_ => v.Result + 1);
		}

		static Task<int> IncOptimalAsync(Task<int> v) {
			return v.ContinueWith(task => task.Result + 1);
		}

		static Task<int> AddAsync(Task<int> a, Task<int> b) {
			return Task.Run(() => {
				return a.Result + b.Result;
			});
		}

		static Task<int> Add2Async(Task<int> a, Task<int> b) {
			return Task.Run(() => {
				Task<int>.WaitAll(a, b);
				return a.Result + b.Result;
			});
		}

		static Task<int> AddAlmostOptimalAsync(Task<int> a, Task<int> b) {
			return Task<int>.Factory.ContinueWhenAll<int>(new [] { a, b }, _ => {
				return a.Result + b.Result;
			});
		}

		static Task<int> AddOptimalAsync(Task<int> a, Task<int> b) {
			return Task<int>.Factory.ContinueWhenAll<int>(new [] { a, b }, tasks => {
				return tasks[0].Result + tasks[1].Result;
			});
		}

		static void Main(string[] args) {
			var a = GetValueAsync(1);
			var b = GetValueAsync45(2);

			var r1 = AddAsync(
				Add2Async(a, GetValueAsync(3)),
				IncAsync(Inc2Async(b))
			);
			var r2 = AddOptimalAsync(
				AddAlmostOptimalAsync(a, GetValueAsync(3)),
				IncAlmostOptimalAsync(IncOptimalAsync(b))
			);

			int taskId = Task.WaitAny(r1, r2);
			if (taskId == 0) {
				PrintAddResult(r1);
				PrintOptimalResult(r2);
			} else {
				PrintOptimalResult(r2);
				PrintAddResult(r1);
			}
		}

		static void PrintAddResult(Task<int> t) {
			Console.WriteLine("Result: {0}", t.Result);
		}

		static void PrintOptimalResult(Task<int> t) {
			Console.WriteLine("Optimal result: {0}", t.Result);
		}
	}
}
